package fi.jyu.mit.Music;

import java.util.List;


/**
 * BasicSoundSystem tarjoaa rajapinnan yksinkertaiselle MIDI-äänentoistolle Java-ohjelmissa
 * niin komentoriviltä kuin graafisilta sovelluksilta.
 * <h2>Yleistä</h2>
 * <p>
 * Rajapinta tarjoaa seuraavat funktiot äänentoistoon:
 * </p>
 * <ul>
 *   <li>play monen nuotin toistoon (esim. play("cccc")</li>
 *   <li>playSingleNote yhden nuotin toistoon</li>
 *   <li>rest taukoja varten</li>
 *   <li>setTempo tempon säätelyyn BPM:n (engl. beats per minute eli iskuja minuutissa, oletus 120)</li>
 * </ul>
 * <p>
 * Yksinkertaiseen äänentoistoon soveltuu parhaiten SoundSystem. BufferedSoundSystem puskuroi saamansa datan
 * ja toistaa sen asynkronisesti eri säikeessä kun taas ConcurrentSoundSystem toistaa ääntä reaaliaikaisesti 
 * sitä mukaa mitä annetaan.
 * </p><p>
 * Lisäksi ConcurrentSoundSystem tarjoaa mahdollisuuden kytkeä toiston suoraan standardisyätteeseen, eli 
 * käyttäjä voi tehdä pianon (tai muun soittimen).
 * </p>
 * @see BufferedMidiPlayer
 * @see MidiPlayer
 * @see ThreadedMidiPlayer
 */        
public interface BasicMidiPlayer {    
    
    /**
     * Soittaa merkkijonon nuotit
     * @param sequence soittettavat nuotit
     */
    public void play(String sequence);
    
    /**
     * Soittaa merkkijonon nuotit
     * @param sequence soittettavat nuotit
     * @param length kunkin nuotin pituus
     */
    public void play(String sequence, double length);
    
    /**
     * Soittaa merkkijonon nuotit
     * @param sequence soittettavat nuotit
     * @param noteLength kunkin nuotin pituus
     * @param octave oktaavi josta soitetaan
     */
    public void play(String sequence, double noteLength, int octave);

    /**
     * Soittaa merkkijonon nuotit
     * @param sequence soittettavat nuotit
     * @param length kunkin nuotin pituus
     * @param octave oktaavi josta soitetaan
     * @param velocity kuinka nopeasti kosketin painetaan
     */
    public void play(final String sequence, final double length, final int octave, final int velocity);

    /**
     * Soittaa yksittäisen nuotin 
     * @param note soitettava nuotti
     */
    public void playSingleNote(String note);
    
    /**
     * Soittaa yksittäisen nuotin 
     * @param note soitettava nuotti
     * @param length nuotin pituus
     */
    public void playSingleNote(String note, double length);
    
    
    /**
     * Soittaa yksittäisen nuotin 
     * @param note soitettava nuotti
     * @param length nuotin pituus
     * @param octave oktaavi josta soitetaan
     */
    public void playSingleNote(String note, double length, int octave);

    
    /**
     * Soittaa yksittäisen nuotin 
     * @param note soitettava nuotti
     * @param length nuotin pituus
     * @param octave oktaavi josta soitetaan
     * @param velocity kuinka nopeasti kosketin painetaan
     */
    public void playSingleNote(String note, double length, int octave, int velocity);

    /**
     * Soitetaan lista nuotteja 
     * @param notes mitkä nuotit soitetaan
     */
    public void play(final List<Note> notes);

    // public void rest();
    
    // public void rest(double duration);
    
    /**
     * @param bpm asetettav tempo lyöntiä minuutissa
     */
    public void setTempo(int bpm);
}
