package fi.jyu.mit.Music;

import java.util.ArrayList;
import java.util.List;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Sequence;
import javax.sound.midi.Sequencer;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Track;

/**
 * Puskuroitu äänijärjestelmä, jonka tarkoitus on tukea mahdollisuutta 
 * syöttää äänijärjestelmälle nuotteja, jotka voidaan toistaa
 * joko pääohjelman säikeessä tai uudessa säikeessä.
 * 
 * Tälläin käskyt play ja playNote eivät toista saamiansa nuotteja heti,
 * vaan ne puskuroidaan, eli varastoidaan. Soittaminen aloitetaan metodeilla
 * run tai runAsync.
 * 
 * Metodilla run äänijärjestelmä aloittaa soittamisen nykyisessä säikeessä,
 * kun taas runAsync tekee uuden säikeen joka sitten pyörii niin kauan kunnes
 * soitto on loppunut.
 * 
 * Soittimessa on useita tarckejä (tai kanaviksikin voi ajatella),
 * joihin voi "soittaa" eri soittimia.  Trackin voi tahdistaa
 * keskenään jos johonkin ei ole pitkän aikaan annettu nuotteja.
 * 
 * Asynkronisen soittamisen voi aina lopettaa käskyllä stop.
 * 
 * @author vesal
 * @version 12.9.2010
 * 
 * TODO: rinnakkain soivat nuotit esim syntaksilla A(FG)A
 *       voidaan tehdä esim. niin, että nuotti voi koostua nuoteista
 *       ja koostetun nuotin getLength palauttaa pisimmän sen sisällä olevan nuotin.
 *       Entä miten voidaan sanoa että ensin alkuun yksi nuotti ja ennen kuin
 *       se loppuu alkaa toinen ja ne  soivat vähän aikaa yhtäaikaa?
 *       Pitäisikö sittenkin sanoa tyyliin: A(-F*+G*-)A eli ensin
 *       soitetaan A, sitten alkaa yhtäaikaisuus jossa toiseen taukoa,
 *       toisessa jo alkaa G, sitten toisessa F ja soittoa jää päällekkein
 *       yhden nuotin verran ja F soi yksin loppuun (G:n perässä olevaa taukoa
 *       ei tarvittaisi oikeastaan) ja sitten kun F loppuu, jatkuu pelkäällä A:lla.
 *       Tällöin pienen kirjoitettu yhtäaikaisuus olisi A(F+G)A  
 * TODO: miettivät onko tempo (bmp) otettu huomioon oikeassa paikassa, pitäisikö
 *       huomioida vasta soittohetkellä?
 * TODO: saako nuotit poistaa soittamisessa?
 * TODO: pitäisikö soittamisen ajaksi tehdä kopio trackeistä?
 * TODO: ASCII-tiedoston soitti niin, että sieltä luetaan bpm, kanavien soittimet
 *       ja kullekin kanavalle tulevat nuotit, esim, syntaksilla 0: CCCA 
 * TODO: pitäisikö syntaksia laajentaa sen verran että voisi sanoa myös
 *       A*3/4 joka olisi nätimpi kuin A*0.75      
 */
public class BufferedMidiPlayer2 extends MidiPlayer {
    
    
    /**
     * Nuottipuskuri. Soitetaan tulojärjestyksessä.
     */
    private List<NoteTrack> tracks;
    
    private Sequencer sequencer = null;
    
    /**Alutsteaan tyhjällä listalla
     * 
     */
    public BufferedMidiPlayer2() {
        tracks = new ArrayList<NoteTrack>();
        tracks.add(new NoteTrack());
    }
    
    
    /**
     * Alustetaan valitulle kanavalle
     * @param channel mille kanavalla alustetaan
     */
    public BufferedMidiPlayer2(int channel) {
        this();
        selectChannel(channel);
    }
    
    
    /**
     * Luo kopion puskuroiduista nuoteista
     * @param index kopioitavan trackin numero
     * @return kopio nuoteista
     */
    public List<Note>getCopyOfNotes(int index) {
        return getTrack(index).getCopyOfNotes();
    }
    
    /**
     * Palauttaa nuotit viitteenä jolloin niiden muuttaminen muuttaa soitettavia nuotteja
     * @param index kopioitavan trackin numero
     * @return nuotit
     */
    public List<Note>getNotes(int index) {
        return getTrack(index).getNotes();
    }
    
    /**
     * Tallentaa nuotin puskuriin.
     */
    @Override
    public void midiSoundNote(int note, int length, int velocity) {
        tracks.get(0).add(new Note(note, length, velocity));
    }
    
    @Override
    protected void rest(final int barLength) {
        tracks.get(0).add(new Note.Delay(barLength));
    }
    
    
    /**
     * Aloittaa soiton rinnakkaisesti ja jatkaa kunnes se loppuu.
     * Soiton loputtua on kutsuttava stop(), muuten ohjelma jää käyntiin. 
     */
    public void runAsync() {
        long mul = 1;
        try {
            Sequence song = new Sequence(0, 4 * 120 ); //* getTempo());
            Track track = song.createTrack();
            for (int channel = 0; channel < tracks.size(); channel++) {
                NoteTrack nt  = tracks.get(channel);
                if (nt == null)     continue;
                if ( nt.isMuted() ) continue;
                long tick = nt.getStartTime()*mul;
                ShortMessage msgInst = new ShortMessage();
                msgInst.setMessage(ShortMessage.PROGRAM_CHANGE, channel, nt.getInstrument() - 1, 0);
                track.add(new MidiEvent(msgInst, tick));
                while (!nt.isEmpty()) {
                    Note n = nt.remove(0);
                    if (n.isDelay()) {
                        tick += mul * n.getLength();
                        message("Tauko         " + n.message());
                    } else {
                        try {
                            ShortMessage msgOn = new ShortMessage();
                            msgOn.setMessage(ShortMessage.NOTE_ON, channel, n.getNote(), n.getVelocity());
                            track.add(new MidiEvent(msgOn, tick));
                            message("Soitin nuotin " + n.message());
                            tick += mul * n.getLength();
                            ShortMessage msgOff = new ShortMessage();
                            msgOff.setMessage(ShortMessage.NOTE_OFF, channel, n.getNote(), n.getVelocity());
                            track.add(new MidiEvent(msgOff, tick));
                        } catch (InvalidMidiDataException e) {
                            //
                        }
                    }
                    
                }
            }
            sequencer = MidiSystem.getSequencer();
            sequencer.open();
            sequencer.setSequence(song);

            if ( isVerbose() ) printSequence(song);

            // Aloita soitto
            sequencer.start();
        } catch (InvalidMidiDataException e) {
            error("Virhe luettaessa MIDI-dataa! Annoithan MIDI-tiedoston? (.mid)");
        } catch (MidiUnavailableException e) {
            error("MIDI-soitinta ei voida alustaa.");
        }

    }
    
    
    /**
     * Pysäyttää asynkronisen soiton.
     */
    public void stop() {
        if ( sequencer != null && sequencer.isRunning() ) {
            sequencer.stop();
        }
        if ( sequencer != null ) sequencer.close();
        sequencer = null;
    }
    
    
    /**
     * Ajaa kaikki trackit niin, että päääohjelmaan palataan vasta kun kaikki soitettu
     */
    public void run()  {
        runAsync();
        while ( sequencer.isRunning() ) {
            try {
                Thread.sleep(200);
            } catch (InterruptedException e) {
                //
            }
        }
        stop();
    }
    
    
    /**
     * Vaihtaa instrumentin. Ks. <a href="http://www.midi.org/techspecs/gm1sound.php#instrument">spesifikaatio</a>.
     * @param instrument Instrumentti väliltä 1-128.
     */
    @Override
    public void setInstrument(final int instrument) {
        setInstrument(0, instrument);
    }
    
    /**
     * Vaihtaa instrumentin. Ks. <a href="http://www.midi.org/techspecs/gm1sound.php#instrument">spesifikaatio</a>.
     * @param index track johon instrumentti vaihdetaan
     * @param instrument Instrumentti väliltä 1-128.
     */
    @Override
    public void setInstrument(final int index, final int instrument) {
        super.setInstrument(instrument);
        getTrack(index).setInstrument(instrument);
    }
    
    
    /**
     * Palauttaa pyydetyn trackin.  Jos ei ole, luodaan
     * @param index mikä track halutaan
     * @return uusi track tai indeksistä löytyvä track. 
     */
    public NoteTrack getTrack(int index) {
        NoteTrack nt;
        if ( index < 0 ) return tracks.get(0);
        if ( index < tracks.size() ) {
            nt = tracks.get(index);
            if ( nt != null ) return nt;
            nt = new NoteTrack();
            tracks.set(index,nt);
            return nt;
        }
        while ( index >= tracks.size() ) tracks.add(null);
        nt = new NoteTrack();
        tracks.set(index,nt);
        return nt;
    }


    /**
     * Soittaa nuotit
     * @param index track jolla soitetaan
     * @param notes soitettavat nuotit
     */
    
    public void play(int index, final List<Note> notes) {
        for (Note n : notes) {
            getTrack(index).add(n);
        }
    }
    
    /**
     * Soittaa nuotit 0-trackille
     * @param notes soitettavat nuotit
     */
    @Override
    public void play(final List<Note> notes) {
        play(0,notes);
    }
    
    
    
   /** 
    * @param index track jolla soitetaan
    * @param sequence Nuotit
    * @param octave Oktaavi
    * @param length Yksittäisen nuotin kesto
    * @param velocity voimakkuus, 64 on normaali
    */
   public void play(int index,final String sequence, final double length, final int octave, final int velocity) {
       play(index, getNotes(sequence, length, octave, velocity));
   }
   
   /** 
    * @param index track jolla soitetaan
    * @param sequence Nuotit
    * @param octave Oktaavi
    * @param length Yksittäisen nuotin kesto
    */
   public void play(int index,final String sequence, final double length, final int octave) {
       play(index,sequence, length, octave, getVelocity());
   }
    
   /** 
    * @param index track jolla soitetaan
    * @param sequence Nuotit
    * @param length Yksittäisen nuotin kesto
    */
   public void play(int index,final String sequence, final double length) {
       play(index,sequence, length, 4, getVelocity());
   }
    
   /** 
    * @param index track jolla soitetaan
    * @param sequence Nuotit
    */
   public void play(int index,final String sequence) {
       play(index,sequence, 0.25, 4, getVelocity());
   }
    
    
   /**
    * Vaimennetaan tietty track pois
    * @param index mikä track pysäytetään
    * @param muted pysäytetäänkä vai ei
    */
   public void mute(int index,boolean muted) {
       getTrack(index).mute(muted);
   }
   
   /**
    * @param index tutkittava track
    * @return true jos vaimennettu, muuten false
    */
   public boolean isMuted(int index) {
       return getTrack(index).isMuted();
   }
   
   /**
    * Palauttaa kavan nykyajanhetken kanavan alusta
    * @param index mikä kanava
    * @return nykyajanhetki
    */
   public int getCurrentTime(int index) {
       return getTrack(index).getCurrentTime();
   }
   
   
   /**
    * Asettaa alkuajan.  Voi käyttää esim jos toinen kavan on jo
    * paljon soittanut ja halutaan aloittaa toinen nykykohdasta.
    * esim:  mp.setStartTime(1, mp.getCurrentTime(0));
    * @param index     minkä kanvan aloitusaika asetetaan
    * @param startTime uusi aloitusaika
    */
   public void setStartTime(int index, int startTime) {
       getTrack(index).setStartTime(startTime);
   }
   
   
   /**
    * Toinen tapa saada kanavat samaan tahtiin.  Lisää tauon, jolla
    * päästään samaan kohtaan. Lisätään taukjo siihen kanavaan, jossa
    * on vähemmän aikaa.
    * @param dest mikä kanava synkronoidaan
    * @param source mihin kanavaan
    */
   public void sync(int dest, int source) {
       int sourceTime = getCurrentTime(source);
       int destTime = getCurrentTime(dest);
       
       int dt = sourceTime - destTime;
       
       if ( dt > 0 ) getTrack(dest).add(new Note.Delay(dt));
       if ( dt < 0 ) getTrack(source).add(new Note.Delay(dt));
   }
   
}
