package fi.jyu.mit.Music;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * Luokka yhdelle ääniraidalle
 * @author vesal
 * @version 12.9.2010
 */
public class NoteTrack implements Iterable<Note> {
        private List<Note> noteBuffer;

        private int instrument = 1;
        private int startTime = 0;
        
        /**
         * @return alkuaika
         */
        public int getStartTime() {
            return startTime;
        }

        /**
         * @return nykyinen aika
         */
        public int getCurrentTime() {
            int t = getStartTime();
            for (Note n:noteBuffer) t += n.getLength();
            return t;
        }

        /**
         * @param startTime asetettava alkuaika
         */
        public void setStartTime(int startTime) {
            this.startTime = startTime;
        }

        private boolean muted = false;
        
        /**
         *  alustetaan track tyhjäksi
         */
        public NoteTrack() {
            noteBuffer = new LinkedList<Note>(); 
        }
        
        
        /**
         * @return trackin soitin
         */
        public int getInstrument() {
            return instrument;
        }
    
        /**
         * @param instrument asetettava soitin trackille
         */
        public void setInstrument(int instrument) {
            this.instrument = instrument; 
        }
        
        /**
         * Luo kopion puskuroiduista nuoteista
         * @return kopio nuoteista
         */
        public List<Note>getCopyOfNotes() {
            List<Note> notes = new LinkedList<Note>();
            for (Note n:noteBuffer)
                notes.add(new Note(n));
            return notes;
        }
        
        /**
         * Palauttaa nuotit viitteenä jolloin niiden muuttaminen muuttaa soitettavia nuotteja
         * @return nuotit
         */
        public List<Note>getNotes() {
            return noteBuffer;
        }
        
        /**
         * Tallentaa nuotin puskuriin.
         * @param note nuotin arvo
         * @param length nuotin kesto
         * @param velocity koskettimen lyönnin nopeus
         */
        public void midiSoundNote(int note, int length, int velocity) {
            noteBuffer.add(new Note(note, length, velocity));
        }
        
        
        /**
         * @param barLength lisättävä tauon pituus
         */
        public void rest(final int barLength) {
            // Ei negatiivisia taukoja
            if (barLength < 0)
                return;
            noteBuffer.add(new Note.Delay(barLength));
        }

        /**
         * @param note listättävä nuotti
         */
        public void add(Note note) {
            noteBuffer.add(note);
        }

        @Override
        public Iterator<Note> iterator() {
            return noteBuffer.iterator(); 
        }
        
        /**
         * @return onko track tyhjä
         */
        public boolean isEmpty() {
            return noteBuffer.isEmpty();
        }
        
        /**
         * @param index mistä paikasta poistetaan
         * @return poistettu nuotti
         */
        public Note remove(int index) {
            return noteBuffer.remove(index);
        }

        /**
         * @param muted1 hlejennetäänkö track
         */
        public void mute(boolean muted1) {
            this.muted  = muted1; 
        }
        
        /**
         * @return onkjo track hiljennetty
         */
        public boolean isMuted() {
            return muted;
        }

        
        @Override
        public String toString() {
            return noteBuffer.toString();
        }
        
        
}
